import numpy as np
import pandas as pd
import operator as op
import math
import csv


class Emblem:
	def __init__(self, filename, sep=',', debug=False):
		self.debug = debug
		# READ INPUT FILENAME ---------------------------------------------------------------------
		self.file = []
		with open(filename, 'rt') as csvfile:
			spamreader = csv.reader(csvfile, quotechar='"', delimiter=sep)
			for row in spamreader:
				self.file.append(np.array(row).astype(str))
		self.file_len = len(self.file)

		# Coefficients ----------------------------------------------------------------------------
		self.base_weight = 0
		self.univariate_beta = {}
		self.single_univariate_beta = {}
		self.bivariate_beta = {}
		self.bivariate_beta_flat = {}
		self.end_bivariates_reached = False

		# Rows ------------------------------------------------------------------------------------
		self.start_univariate_row = 0
		self.last_univariate_row = 0

		# Variables ------------------------------------------------------------------------------------
		self.univariate_variables = []
		self.bivariate_variables = []
		self.read_base_level()
		self.read_univariates()
		self.read_bivariates()
		self.flatten_bivariates()

	# BASE LEVEL ----------------------------------------------------------------------------------
	def read_base_level(self):

		self.base_weight = float(self.file[0][2])
		self.start_univariate_row = 5

	# UNIVARIATES ----------------------------------------------------------------------------------
	def read_univariates(self):
		univariate_levels_cols = np.array(np.where(self.file[self.start_univariate_row] != '')[0])
		self.univariate_beta = {}
		self.last_univariate_row = 0

		for col in univariate_levels_cols:
			self.single_univariate_beta = {}
			row = self.start_univariate_row + 1
			single_univariate_end_reached = False
			univariate_name = self.file[row - 1][col]
			if self.debug:
				print('{:-<100}'.format('{} '.format(univariate_name)))

			while not single_univariate_end_reached:

				if self.file_len == row + 1:
					single_univariate_end_reached = True
					self.end_bivariates_reached = True
				elif self.file[row + 1][col + 1] == '':
					single_univariate_end_reached = True
				try:
					level = float(self.file[row][col])
				except ValueError:
					level = self.file[row][col]

				if self.debug:
					print('\trow: {:<6}| col: {:<6}| level: {:<40}| value: {:<8}'.format(row, col, level, self.file[row][col + 1]))

				self.single_univariate_beta.update({level: float(self.file[row][col + 1])})
				row += 1

			self.univariate_beta.update({univariate_name: self.single_univariate_beta})
			self.last_univariate_row = max(self.last_univariate_row, row)

			self.univariate_variables = list(self.univariate_beta.keys())

	def read_bivariates(self):
		row = self.last_univariate_row + 2
		self.bivariate_beta = {}

		if not self.end_bivariates_reached:
			if not (self.file[row - 2][0] == 'Orthogonal Polynomial Equations' or self.file[row - 1][0] == 'Orthogonal Polynomial Equations' or self.file[row][0] == 'Orthogonal Polynomial Equations'):
				# cycle over bivariate matrices
				while not self.end_bivariates_reached:
					first_bivariate_name = self.file[row + 2][0]
					second_bivariate_name = self.file[row][2]
					bivariate_interaction_name = first_bivariate_name + '|' + second_bivariate_name

					try:
						second_bivariate_levels = list(map(float, self.file[row + 1][np.where(self.file[row + 1] != '')]))
					except ValueError:
						second_bivariate_levels = self.file[row + 1][np.where(self.file[row + 1] != '')]

					row += 2

					# cycle over bivariate matrix's rows
					single_bivariate_end_reached = False
					single_bivariate_beta = {}
					while not single_bivariate_end_reached:
						try:
							first_bivariate_level = float(self.file[row][1])
						except ValueError:
							first_bivariate_level = self.file[row][1]
						first_bivariate_level = first_bivariate_level if str(first_bivariate_level) != '' else 'nan'

						single_bivariate_beta.update({first_bivariate_level: dict(zip(second_bivariate_levels, list(map(float, self.file[row][2:2 + len(second_bivariate_levels)]))))})

						if row >= self.file_len - 1:
							single_bivariate_end_reached = True
							self.end_bivariates_reached = True
						elif self.file[row + 1][2] == '':
							single_bivariate_end_reached = True

							if row < self.file_len - 1:
								if self.file[row + 2][0] == 'Orthogonal Polynomial Equations':
									self.end_bivariates_reached = True

						row += 1

					row += 3

					self.bivariate_beta.update({bivariate_interaction_name: single_bivariate_beta})

				self.bivariate_variables = list(set(np.array(list(map(lambda x: x.split('|'), self.bivariate_beta.keys()))).flatten()))

	def flatten_bivariates(self):
		self.bivariate_beta_flat = {}
		for interaction_name, bivariate_dict in self.bivariate_beta.items():
			flat_dict = pd.DataFrame(bivariate_dict).unstack()
			self.bivariate_beta_flat.update({interaction_name: {'{}|{}'.format(k[0], k[1]): v for k, v in flat_dict.items()}})

	def linear_predictor(self, df: pd.DataFrame, excluded_vars=None, debug=False) -> np.array:
		return self.evaluate(df=df, operator=op.add, excluded_vars=excluded_vars, debug=debug)

	def evaluate(self, df: pd.DataFrame, operator, excluded_vars: list=None, debug=False) -> np.array:
		if excluded_vars is None:
			excluded_vars = []

		df = df.reset_index(drop=True)
		result = np.array([self.base_weight for _ in range(len(df))])

		for variable, beta_dict in self.univariate_beta.items():
			if variable not in excluded_vars:
				if debug:
					try:
						result = operator(result, df[variable].apply(lambda x: beta_dict[x]))
					except KeyError as ke:
						print(variable, ke)
						raise KeyError
				else:
					result = operator(result, df[variable].map(beta_dict))

		for interaction_name, interaction_dict in self.bivariate_beta_flat.items():
			first_bivariate = interaction_name.split('|')[0]
			second_bivariate = interaction_name.split('|')[1]

			first_bivariate_values = df[first_bivariate]
			second_bivariate_values = df[second_bivariate]

			if not ((first_bivariate in excluded_vars) or (second_bivariate in excluded_vars)):
				try:
					first_bivariate_values = np.round(first_bivariate_values.astype(float), 3)
				except ValueError:
					pass

				try:
					second_bivariate_values = np.round(second_bivariate_values.astype(float), 3)
				except ValueError:
					pass

				compact_bivariate = first_bivariate_values.astype(str) + pd.Series(['|' for _ in range(len(df))]) + second_bivariate_values.astype(str)

				if debug:
					try:
						result = operator(result, compact_bivariate.apply(lambda x: interaction_dict[x]))
					except KeyError as ke:
						try:
							ke = float(str(ke))
						except TypeError:
							pass

						if math.isnan(ke):
							error_condition = compact_bivariate.apply(lambda x: math.isnan(x) if isinstance(x, float) else False)
						else:
							error_condition = compact_bivariate.isin([ke])

						print(interaction_name, ke, '\n')
						print(pd.concat([first_bivariate_values, second_bivariate_values, compact_bivariate], axis=1)[error_condition].iloc[0:10, :])
						raise KeyError
				else:
					result = operator(result, compact_bivariate.map(interaction_dict))

		null_values_index = result.isna()
		null_values = null_values_index.sum()

		if result.isna().sum() > 0:
			print('Warning! {} null values found in result'.format(null_values))

		return np.array(result.reset_index(drop=True))

	def predict(self, df: pd.DataFrame, link_function, excluded_vars=None, debug=False) -> pd.Series:
		link_function = np.vectorize(link_function)
		return pd.Series(data=link_function(self.linear_predictor(df, excluded_vars, debug)), index=df.index)

	def get_beta(self, df: pd.Series, index_variable=None):
		beta_values = []
		for idx, row in df.iterrows():
			beta_values_tmp = {
				'index': idx if index_variable is None else row[index_variable],
				'base_weight': self.base_weight,
				'weights': {}
			}
			for variable, beta_dict in self.univariate_beta.items():
				beta_values_tmp['weights'].update({variable: {row[variable]: beta_dict[row[variable]]}})

			for interaction_name, interaction_dict in self.bivariate_beta.items():
				first_bivariate = interaction_name.split('|')[0]
				second_bivariate = interaction_name.split('|')[1]
				beta_values_tmp['weights'].update({
					'{}|{}'.format(first_bivariate, second_bivariate): {'{}|{}'.format(row[first_bivariate], row[second_bivariate]): interaction_dict[row[first_bivariate]][row[second_bivariate]]}
				})

			beta_values.append([beta_values_tmp])

		return pd.DataFrame(data=beta_values, index=df.index, columns=['beta_values'])



import pandas
import numpy
import re
from scipy import stats

from collections import Counter


def take_last(s: pandas.Series):
	return s.iloc[-1]


def take_first(s: pandas.Series):
	return s.iloc[0]


def count_unique(x: pandas.Series):
	return len(x.unique())


def bin_variable(s: pandas.Series, min_value, max_value, width, how, default=None, lower=None, upper=None, errors=None):
	"""
	bin pandas Series
	:param s: pandas Series
	:param min_value: minimum bining value
	:param max_value: maximum binning value
	:param width: bin width
	:param how: function to use for binning. Use numpy.ceil, numpy.floor
	:param default: default value to replace nans
	:param lower: value to assign to values smaller than min_value
	:param upper: value to assign to values larger  than max_value
	:param errors: error values in the Series to be treated a if they where nans
	:return: binned series s_binned
	"""
	s_binned = s.copy()
	nan_values = numpy.array(s_binned.isna()) | numpy.array((s_binned == errors))

	conditions = numpy.array([True for _ in range(len(s_binned))])

	if lower is not None:
		s_binned.loc[s_binned < min_value] = lower
		conditions = conditions & (s_binned != lower) & (min_value <= s_binned)

	if upper is not None:
		s_binned.loc[s_binned > max_value] = upper
		conditions = conditions & (s_binned != upper) & (s_binned <= max_value)

	s_binned.loc[conditions] = list(numpy.clip(how(s_binned.loc[conditions]/width) * width, min_value, max_value))
	s_binned = numpy.round(s_binned, 8)
	if default is not None:
		s_binned.loc[nan_values] = default

	return s_binned


def detect_outliers(df, n, features):
	"""
	Takes a dataframe df of features and returns a list of the indices
	corresponding to the observations containing more than n outliers according
	to the Tukey method.
	"""
	outlier_indices = []

	# iterate over features(columns)
	for col in features:
		# 1st quartile (25%)
		q1 = numpy.percentile(df[col].dropna(), 25)
		# 3rd quartile (75%)
		q3 = numpy.percentile(df[col].dropna(), 75)
		# Interquartile range (iqr)
		iqr = q3 - q1

		# outlier step
		outlier_step = 1.5 * iqr

		# Determine a list of indices of outliers for feature col
		outlier_list_col = df[(df[col] < q1 - outlier_step) | (df[col] > q3 + outlier_step)].index

		# append the found outlier indices for col to the list of outlier indices
		outlier_indices.extend(outlier_list_col)

	# select observations containing more than 2 outliers
	outlier_indices = Counter(outlier_indices)
	multiple_outliers = list(k for k, v in outlier_indices.items() if v > n)

	return multiple_outliers


def multiple_replace(s: pandas.Series, replacing_dict: dict, default=None):
	"""
	replaces multiple values at once
	:param s: pandas Series to act on
	:param replacing_dict: dictionary with replacements. Has to be in the form: {'<replacing_value_1>' : [<list_of_values_to_replace], '<replacing_value_2>' : [<list_of_values_to_replace] ...}
	:param default: default value. used to fill values not in replacing_value_n
	:return:
	"""
	for key, value in replacing_dict.items():
		s = s.replace(value, key)
	if default is not None:
		s = s.map(dict(zip(replacing_dict.keys(), replacing_dict.keys()))).fillna(default)
	return s


def find_regex(regex, value):
	if re.search(regex, value) is not None:
		return True
	else:
		return False


def fill_empty_with_distr(s: pandas.Series, values_to_fill: list, values_to_avoid: list=None):
	"""
	:param s: pandas Series to be filled
	:param values_to_fill: value to fill
	:param values_to_avoid: values not valid for filling
	:return:
	"""

	compl_values_to_avoid = [numpy.nan]
	if values_to_avoid is not None:
		compl_values_to_avoid += values_to_avoid

	filling_index = s.index[s.isin(values_to_fill)]

	distrib_data = s.value_counts()
	distrib_data = distrib_data[~distrib_data.index.isin(values_to_fill + compl_values_to_avoid)]
	distrib_data_dict = {k: v for k, v in enumerate(distrib_data.index)}

	xk, pk = list(distrib_data_dict.keys()), distrib_data.values/distrib_data.values.sum()
	distrib = stats.rv_discrete(values=(xk, pk))

	filled_s = s.copy()
	filled_s[filling_index] = [distrib_data_dict[distrib.rvs()] for _ in filling_index]

	return filled_s



import numpy
import pandas
from sklearn.feature_extraction import FeatureHasher


class Hasher:
	def __init__(self, categoricals: list, n_min_features: int=2, n_attempt_max: int=3, n_attempt_tot_max: int=30, verbose: bool=False):
		self.categoricals = categoricals
		self.hashing_dict = {}
		self.n_min_features = n_min_features
		self.n_attempt_max = n_attempt_max
		self.n_attempt_tot_max = n_attempt_tot_max
		self.__fitted = False
		self.verbose = verbose

	def fit(self, df: pandas.DataFrame):
		self.__fitted = False

		for variable in self.categoricals:
			if self.verbose:
				print('Fitting', variable)
			self.__get_hashing(df, variable)

		self.__fitted = True

	def transform(self, data, array_indices=None):
		if not self.__fitted:
			raise FitError('Hasher class not fitted')

		is_df = isinstance(data, pandas.DataFrame)
		is_series = isinstance(data, pandas.Series)
		is_dict = isinstance(data, dict)
		is_array = isinstance(data, numpy.ndarray)

		if is_df:
			transformed = data[data.columns.difference(self.categoricals)]
			original_columns = list(data.columns)
		elif is_series:
			transformed = data[data.index.difference(self.categoricals)]
			original_columns = list(data.index)
		elif is_dict:
			transformed = data.copy()
			original_columns = list(data.keys())
		elif is_array:
			if array_indices is None:
				raise MissingIndicesError('Missing indices for array transformation')
			transformed = numpy.array([])
			original_columns = []
		else:
			raise TypeError('Type not valid for transformation')

		if is_df or is_series:
			new_columns = []
			for variable in original_columns:
				if variable in self.categoricals:
					n_new_columns = len(list(self.hashing_dict[variable].values())[0])
					hashing_columns = [variable + '_H' + str(i) for i in range(n_new_columns)]
					new_columns += hashing_columns

					if self.verbose:
						print('Transforming', variable)
					if is_series:
						transformed = transformed.append(pandas.Series(self.hashing_dict[variable][data[variable]], index=hashing_columns))
					else:
						transformed = transformed.join(pandas.DataFrame(list(data[variable].apply(lambda x: self.hashing_dict[variable][x])), index=data.index, columns=hashing_columns))
				else:
					new_columns.append(variable)

			transformed = transformed[new_columns]
			transformed = transformed[new_columns]

		elif is_dict:
			for variable in self.categoricals:
				if self.verbose:
					print('Transforming', variable)

				n_new_columns = len(list(self.hashing_dict[variable].values())[0])
				columns = [variable + '_H' + str(i) for i in range(n_new_columns)]
				value = transformed.pop(variable)
				try:
					hashing_values = self.hashing_dict[variable][value]
				except KeyError:
					raise HashingError('Value \'{}\' not found in hashing dict for variable \'{}\'.'.format(value, variable))

				transformed.update({columns[i]: float(v) for i, v in enumerate(hashing_values)})

		elif is_array:
			for index in range(len(data)):
				if index in array_indices.keys():
					variable = array_indices[index]
					try:
						transformed = numpy.append(transformed, numpy.array(self.hashing_dict[variable][data[index]]))
					except KeyError:
						raise HashingError('Value \'{}\' not found in hashing dict for variable \'{}\'.'.format(data[index], variable))
				else:
					numpy.append(transformed, data[index])
		else:
			raise TypeError('Type not valid for transformation')

		return transformed

	def fit_transform(self, df: pandas.DataFrame):
		self.fit(df)
		return self.transform(df)

	def __get_hashing(self, df: pandas.DataFrame, variable: str):

		n_features = self.n_min_features
		n_attempt = 0
		n_attempt_tot = 0
		lost = -1
		non_trivial = False
		uniques = df.loc[:, variable].unique()
		hashed_matrix = [[]]

		while (lost != 0 or not non_trivial) and n_attempt_tot < self.n_attempt_tot_max:
			non_trivial = False
			feature_hasher = FeatureHasher(input_type='string', n_features=n_features)
			keys_list = list(map(lambda x: str(x) + str(numpy.random.rand())[2:8], uniques))

			hashed_matrix = feature_hasher.fit_transform(keys_list).toarray().astype(int)
			hashed_code = set(map(lambda x: ''.join(x.astype(str)), hashed_matrix))

			lost = len(hashed_code) - len(uniques)

			n_attempt += 1
			n_attempt_tot += 1

			if n_attempt == self.n_attempt_max:
				n_attempt = 0
				n_features += 1

			if min(numpy.var(hashed_matrix, axis=0)) > 0:
				non_trivial = True

		if lost != 0 or not non_trivial:
			raise HashingError('Hashing failed on variable: {}. Lost: {}. Trivial: {}.'.format(variable, lost, not non_trivial))

		else:
			tmp_hashing_dict = dict(zip(uniques, map(list, hashed_matrix)))
			self.hashing_dict.update({variable: tmp_hashing_dict})


class HashingError(Exception):
	""":raises Error when hashing is invalid"""


class FitError(Exception):
	""":raises Error when class Hasher is not fitted"""


class MissingIndicesError(Exception):
	""":raises Error when indices are missing"""



import pandas


class DataReader:
	def __init__(self, default_values_path):
		self.process = None

		self.input_data = {}
		self.input_data_processed = {}
		self.suppress_calculation = False

		self.warnings = []
		self.errors = []

		# default values --------------------------------------------------------------------------
		self.variables_set_to_default = []
		with open(default_values_path, 'r') as f:
			self.default_values = {k: v for (k, v) in [x.replace('\n', '').split(',') for x in f.readlines()[1:]]}

		for variable, default_value in self.default_values.items():
			try:
				new_default_value = float(default_value)
			except ValueError:
				pass
			else:
				if int(new_default_value) == new_default_value:
					new_default_value = int(new_default_value)
				self.default_values.update({variable: new_default_value})

	def message_key_error(self, error: KeyError, additional_message: str = '') -> str:
		return 'KeyError ({}): missing key {} in input_data dict. Setting default value.{}'.format(self.process, error, additional_message)

	def message_key_error_crucial(self, error: KeyError) -> str:
		return 'KeyError ({}): missing crucial key {} in input_data dict. Calculation suppressed.'.format(self.process, error)

	def message_value_error(self, variable: str, wrong_value, additional_message: str = '') -> str:
		return 'ValueError ({}): wrong value \'{}\' for variable \'{}\'. Setting default value. {}'.format(self.process, wrong_value, variable, additional_message)

	def message_type_error(self, variable: str, wrong_value, additional_message: str = '') -> str:
		return 'TypeError ({}): wrong type \'{}\' for variable \'{}\'. Setting default value. {}'.format(self.process, type(wrong_value), variable, additional_message)

	def message_choice_error(self, variable, wrong_value, additional_message: str = ''):
		return 'ChoiceError ({}): value \'{}\' not found in possible choices for variable \'{}\'. Setting default value. {}'.format(self.process, wrong_value, variable, additional_message)

	def message_attribute_error(self, variable, error, additional_message: str = ''):
		return 'AttributeError ({}): {}. Variable \'{}\'. Setting default value. {}'.format(self.process, error, variable, additional_message)

	def handle_exceptions(self, exceptions_handler: list, variable: str, error: Exception, wrong_value, additional_message: str = None):
		if isinstance(error, ValueError):
			exceptions_handler.append(self.message_value_error(variable=variable, wrong_value=wrong_value, additional_message=additional_message))
		elif isinstance(error, TypeError):
			exceptions_handler.append(self.message_type_error(variable=variable, wrong_value=wrong_value, additional_message=additional_message))
		else:
			exceptions_handler.append('Unhandled exception thrown: {}'.format(error))

	def get_value(self, variable: str, exceptions_handler: list, is_crucial: bool = False, get_error: bool = False, new_name=None, set_default=True):
		name = variable if new_name is None else new_name
		is_error = False
		value = None

		try:
			value = self.input_data[variable]
		except KeyError as kE:
			is_error = True

			if is_crucial:
				exceptions_handler.append(self.message_key_error_crucial(error=kE))
				self.suppress_calculation = True
			else:
				if set_default:
					value = self.set_default_value(name)
				exceptions_handler.append(self.message_key_error(error=kE))

		if get_error:
			return value, is_error
		else:
			return value

	def get_str(self, variable: str, exceptions_handler: list, new_name: str = None, set_lowercase: bool = False, is_crucial: bool = False):
		name = variable if new_name is None else new_name
		value, is_error = self.get_value(variable=variable, exceptions_handler=exceptions_handler, is_crucial=is_crucial, get_error=True, new_name=new_name)

		if not is_error:
			try:
				value = str(value)
			except (ValueError, TypeError) as error:
				self.handle_exceptions(exceptions_handler=exceptions_handler, variable=variable, error=error, wrong_value=value)

				if is_crucial:
					value = None
					self.suppress_calculation = True
				else:
					value = self.set_default_value(name)
			else:
				if set_lowercase:
					value = value.lower()

		return value

	def set_str(self, variable: str, exceptions_handler: list, new_name: str = None, set_lowercase: bool = False, is_crucial: bool = False):
		name = variable if new_name is None else new_name
		value = self.get_str(variable, exceptions_handler, new_name, set_lowercase, is_crucial)
		self.input_data_processed.update({name: value})

	def get_categorical(self, variable: str, choices: list, exceptions_handler: list, force_type=None, new_name: str = None, set_lowercase: bool = False, is_crucial: bool = False):
		name = variable if new_name is None else new_name
		value, is_error = self.get_value(variable=variable, exceptions_handler=exceptions_handler, is_crucial=is_crucial, get_error=True, new_name=new_name)

		if not is_error:
			try:
				if set_lowercase:
					value = str(value).lower()
				if force_type is not None:
					value = force_type(value)
			except ValueError:
				is_error = True
				exceptions_handler.append('ValueError ({}): impossible to force type {} on {}. Setting default value {}'.format(self.process, force_type, self.input_data[variable], value))
			if value not in choices:
				is_error = True
				exceptions_handler.append(self.message_choice_error(variable=name, wrong_value=value))

			if is_error:
				if is_crucial:
					value = None
					self.suppress_calculation = True
				else:
					value = self.set_default_value(name)

		return value

	def set_categorical(self, variable: str, choices: list, exceptions_handler: list, force_type=None, new_name: str = None, set_lowercase: bool = False, is_crucial: bool = False):
		name = variable if new_name is None else new_name
		value = self.get_categorical(variable, choices, exceptions_handler, force_type, new_name, set_lowercase, is_crucial)
		self.input_data_processed.update({name: value})

	def get_numeric(self, variable: str, exceptions_handler: list, numeric_type: type = int, min_value: int = None, max_value: int = None, bucket: int = None, new_name: str = None, is_crucial=False):
		name = variable if new_name is None else new_name
		value, is_error = self.get_value(variable=variable, exceptions_handler=exceptions_handler, is_crucial=is_crucial, get_error=True, new_name=new_name)

		if not is_error:
			try:
				value = numeric_type(value)
			except (ValueError, TypeError) as error:
				is_error = True
				self.handle_exceptions(exceptions_handler=exceptions_handler, variable=name, error=error, wrong_value=value)
			else:
				if value != bucket:
					if min_value is not None:
						value = max(min_value, value)
					if max_value is not None:
						value = min(max_value, value)
			if is_error:
				if is_crucial:
					value = None
					self.suppress_calculation = True
				else:
					value = self.set_default_value(name)

		return value

	def set_numeric(self, variable: str, exceptions_handler: list, numeric_type: type = int, min_value: int = None, max_value: int = None, bucket: int = None, new_name: str = None, is_crucial=False):
		name = variable if new_name is None else new_name
		value = self.get_numeric(variable, exceptions_handler, numeric_type, min_value, max_value, bucket, new_name, is_crucial)
		self.input_data_processed.update({name: value})

	def get_date(self, variable: str, exceptions_handler: list, new_name: str=None, is_crucial: bool=False, timezone: str='Europe/London'):
		name = variable if new_name is None else new_name
		date, is_error = self.get_value(variable=variable, exceptions_handler=exceptions_handler, is_crucial=is_crucial, get_error=True, new_name=new_name)

		if not is_error:
			try:
				date = pandas.to_datetime(date)
				if date is None:
					raise ValueError
			except (ValueError, AttributeError) as error:
				is_error = True
				if isinstance(error, ValueError):
					self.warnings.append(self.message_value_error(variable=name, wrong_value=date))
				if isinstance(error, AttributeError):
					self.warnings.append(self.message_attribute_error(variable=name, error=date))

			if is_error:
				if is_crucial:
					date = None
					self.suppress_calculation = True
				else:
					date = self.set_default_value(name)
			else:
				date = date.tz_convert(timezone)

		return date

	def set_date(self, variable: str, exceptions_handler: list, new_name: str = None, is_crucial=False, timezone: str='Europe/London'):
		name = variable if new_name is None else new_name
		date = self.get_date(variable=variable, exceptions_handler=exceptions_handler, new_name=new_name, is_crucial=is_crucial, timezone=timezone)
		self.input_data_processed.update({name: date})

	def set_default_value(self, variable):
		try:
			default_value = self.default_values[variable]
		except KeyError:
			raise KeyError('Missing default value for variable {}'.format(variable))
		else:
			self.variables_set_to_default.append(variable)

		return default_value

	def reset(self):
		self.process = None

		self.input_data = {}
		self.input_data_processed = {}
		self.variables_set_to_default = []

		self.warnings = []
		self.errors = []

		self.suppress_calculation = False



import numpy
import pandas


class Scaler:
	def __init__(self, scaler):
		self.__scaler = scaler
		self.__fitted = False
		self.__variables = None

	def fit(self, df: pandas.DataFrame, variables):
		self.__variables = variables
		self.__scaler.fit(df[self.__variables])
		self.__fitted = True

	def transform(self, df):
		if self.__fitted:
			original_variables = list(df.columns)
			other_variables = list(df.columns.difference(self.__variables))
			transformed = self.__scaler.transform(df[self.__variables])

			transformed = df[other_variables].join(transformed)
		else:
			raise Exception('Scaler not fitted')

		return transformed[original_variables]

	def fit_transform(self, df, variables):
		self.fit(df, variables)
		return self.transform(df)


class MultiDimStandardScaler:
	def __init__(self):
		self._mean = None
		self._var = None
		self._std = None
		self.__fitted = None

	def fit(self, df):
		if isinstance(df, pandas.DataFrame):
			data = df.values
		else:
			data = df.copy()

		self._mean = numpy.mean(data, axis=0)
		self._var = numpy.var(data, axis=0)
		self._std = numpy.sqrt(self._var)
		self.__fitted = True

	def transform(self, df):
		if self.__fitted:
			sample_len = len(df)
			scaled = df - numpy.array([self._mean for _ in range(sample_len)])
			scaled = scaled / numpy.array([self._std for _ in range(sample_len)])
		else:
			raise Exception('Scaler not fitted.')

		return scaled

	def fit_transform(self, df):
		self.fit(df)
		return self.transform(df)
